#include <nostub.h>
#include <all.h>
#include "doorsos.h"

#include "kernel_b.h"	// code du kernel


int _ti89, _ti92plus;

char calc = 0;

ulong ROM_table_addresses[NB_ROM_VERSIONS] = {
	 0x478A74,    // TI92+ 1.00
	 0x279578,    // TI89 1.00
	 0x478BB4,    // TI92+ 1.01
	 0x278EAC,    // TI89 1.05
	 0x478280,    // TI92+ 1.05
	 0x279E02,    // TI89 2.01
 	 0x479B0E,    // TI92+ 2.01
	 0x279F3A,    // TI89 2.03
	 0x479C42,    // TI92+ 2.03
     0x23A17A,	  // TI89 2.04
     0x43A0AE};	  // TI92+ 2.04

ushort HeapValues[NB_ROM_VERSIONS] = {
	0x7592,	   // 92+ 1.00
	0x758E,	   // 89  1.00
	0x7592,	   // 92+ 1.01
	0x7596,    // 89  1.05
	0x759A,    // 92+ 1.05
	0x7992,    // 89  2.01
	0x7992,    // 92+ 2.01
	0x7996,    // 89 2.03
	0x7996,    // 92+ 2.03
    0x7966,	   // 89 2.04
    0x7962};   // 92+ 2.04

ushort ROMversions[NB_ROM_VERSIONS] = {
	0x1100,    // 92+ 1.00
	0x0100,    // 89  1.00
	0x1101,    // 92+ 1.01
	0x0105,    // 89  1.05
	0x1105,    // 92+ 1.05
	0x0201,    // 89  2.01
	0x1201,    // 92+ 2.01
	0x0203,    // 89  2.03
	0x1203,    // 92+ 2.03
    0x0204,	   // 89  2.04
    0x1204};   // 92+ 2.04

void display(char *msg)
{
  DlgMessage("DoorsOS II 0.97", msg, BT_OK, BT_NONE);
}

void uninstall()
{
  ulong **rom_int_ptr, **ram_int_ptr;
  int i;
  ushort kernel_bloc = *KERNEL_BLOCK;
  OSSetSR(0x700);	// ints off
  if (kernel_bloc)	// unsinstallation 'new way'
  {
    HeapFree(kernel_bloc);
    EV_hook = NULL;
  }
  else				// old style uninstallation
  {
	asm("
	    move.l	(0xC8),%a4
	    add.l	#0x2F*4,%a4
	    move.l	(%a4),%a4
	    move.l	#0x3fff4,0x10E(%a4)");
  }

   	// Re installs the old interrupts
  rom_int_ptr = *RESET_VECTOR > 0x400000 ? VECTOR_TABLE92 + 0x64/4 : VECTOR_TABLE89 + 0x64/4;
  ram_int_ptr = INTERRUPT_TABLE + 0x64/4;	// Auto int 1
  i = 7;									// 7 auto ints
  while (i--)
    *ram_int_ptr++ = *rom_int_ptr++;
  memset(INTERRUPT_TABLE + 0x30/4, 0, 0x1A);	// [0x40030..0x4004A] = 0

  OSSetSR(0);	// ints on
}

inline unsigned long GetHardwareVersion() // returns the Hardware version
{
  unsigned long hwpb, *rombase;
  rombase = (unsigned long *)((*(unsigned long *)0xC8) & 0x600000);
  hwpb = rombase[65];
  return (hwpb - (unsigned long)rombase < 0x10000 &&
    *(unsigned short *)hwpb > 22 ? *(unsigned long *)(hwpb + 22) : 1);
}

inline uchar IsPatchInstalled()  // Supposing it is a HW2 AMS 2.0x calculator
{
  ulong *ptr;
  if (*(ulong *)0xAC == 0x100)
    return 1;
  ptr = (ulong *)(((void *)EX_stoBCD)+0x52);
  if (*ptr++ != 0x700000)
    return 1;
  if (*ptr != 0x36BCFFDF)
    return 1;
  return 0;
}

void _main()
{
  int rom;
  ushort *ker_ver = KERNEL_VERSION;
  ushort *ker_name = KERNEL_ID;
  ushort *kernel_words = (ushort *)kernel_bin;

  ushort kernel_bloc;
  void *kernel_address;
  ushort kernel_size;

  ulong *RAM_CALLS;
  ulong *rom_ptr;
  ushort handle;

  ushort *reloc_word;	// pour le relogement du kernel
  ushort reloc_offset;
  ushort reloc_value;
  HSym sym;

  calc = *RESET_VECTOR > 0x400000 ? 1 : 0;
  if (*ker_ver)       // un kernel est dja install
  {
    if (*ker_name != DOORSOS_ID) // si ce kernel n'est pas doorsos
      uninstall();
    else
    {
      if (*ker_ver < DOORSOS_VERSION)
       uninstall();
      if (*ker_ver == DOORSOS_VERSION)
      {
        display("Already installed");
        return;
      }
      if (*ker_ver > DOORSOS_VERSION)
      {
        display("Newer version already installed");
        return;
      }
    }
  }

  kernel_size = kernel_words[0]-1;
  kernel_bloc = HeapAllocHigh(kernel_size);
  if (!kernel_bloc)
  {
    display("Not enough memory to install");
    return;
  }
  kernel_address = HeapDeref(kernel_bloc);
  OSSetSR(0x700);

// rom version
  rom = NB_ROM_VERSIONS;
  while (rom--)
    if (*ROM_TABLE == ROM_table_addresses[rom])
      break;
  if ( (rom >= 5) && (GetHardwareVersion() != 1))		// HW2 calcs, AMS >= 2.xx
  {
    if (!IsPatchInstalled())
    {
      DlgMessage("HW2 calculator detected","You MUST install HWPatch 2.10 before DoorsOS", BT_OK, BT_NONE);
      HeapFree(kernel_bloc);
      return;
    }
  }

  *KERNEL_BLOCK = kernel_bloc;	// sauve le bloc du kernel
  *KERNEL_VERSION = DOORSOS_VERSION;
  *KERNEL_ID = DOORSOS_ID;
  *KERNEL_EXEC = kernel_words[1]+kernel_address;
  *KERNEL_RELOC = kernel_words[2]+kernel_address;
  *KERNEL_RELOC2 = kernel_words[3]+kernel_address;
  *KERNEL_UNRELOC = kernel_words[4]+kernel_address;
  *KERNEL_UNRELOC2 = kernel_words[5]+kernel_address;
  EV_hook = kernel_words[6]+kernel_address;
  *LINE_1010 = kernel_words[8]+kernel_address;

// relogement du kernel
  reloc_word = (ushort *)(kernel_bin+kernel_size+2);
  while (*(--reloc_word))
  {
    reloc_offset = *reloc_word;
    reloc_value = *--reloc_word;
    *(ulong*)(kernel_bin+2+reloc_offset) = (ulong)(reloc_value+kernel_address);
  }
  memcpy(kernel_address, kernel_bin+2, kernel_size);	// copie le kernel

// detecter les ram calls...
  RAM_CALLS = *(ulong **)(kernel_words[7]+kernel_bin+2);

// font_medium : thanx to Fred
  rom_ptr = (calc == 1) ? (ulong *)0x400000 : (ulong *)0x200000;
  while (*rom_ptr != 0x54A854A8)
    rom_ptr++;
  RAM_CALLS[FONTS] = (ulong)rom_ptr;
  RAM_CALLS[FONTS+1] = (ulong)rom_ptr;

// kb_vars : Thanx to olivier lesteven
  rom_ptr = (ulong *)&EV_hook;
  while (*rom_ptr != 2)
    rom_ptr = (ulong *) ((ushort*)(rom_ptr)+1);
  rom_ptr -= 6;	// - 6 longwords
  RAM_CALLS[KB_VARS] = (ulong)rom_ptr;
  RAM_CALLS[KB_VARS+1] = (ulong)rom_ptr;

// MainFolderHandle & FolderListHandle
  sym = SymFindHome("\0main"+5);
  handle = sym.folder;	// FolderListHandle
  RAM_CALLS[FOLDERLIST] = (long)handle;
  RAM_CALLS[FOLDERLIST+1] = (long)handle;

  handle = DerefSym(sym)->handle; // MainFolderHandle
  RAM_CALLS[MAINHANDLE] = (long)handle;
  RAM_CALLS[MAINHANDLE+1] = (long)handle;

  OSSetSR(0);

  if (rom<0)
  {
    ulong rom_ver;
    DlgMessage("Warning !", "This AMS version is unknown to DoorsOS, so it may be incompatible", BT_OK, BT_NONE);
    RAM_CALLS[HEAP+1] = RAM_CALLS[HEAP] = (*(ulong *)0xC8)+0x1104;	// HeapTable
    rom_ver = calc ? (ulong)ROMversions[NB_ROM_VERSIONS-1] : (ulong)ROMversions[NB_ROM_VERSIONS-2];
    rom_ver++;
    RAM_CALLS[ROM_VER+1] = rom_ver;
    RAM_CALLS[ROM_VER] = rom_ver;
  }
  else
  {
    RAM_CALLS[HEAP] = HeapValues[rom];     	// Heap
    RAM_CALLS[HEAP+1] = HeapValues[rom];

    RAM_CALLS[ROM_VER] = ROMversions[rom]; 	// rom version
    RAM_CALLS[ROM_VER+1] = ROMversions[rom];
  }
  display("Installation successful");
}


