#define EXTRA_RAM	0x4000
#define WORD_REF	0x8000

char calculator = 0;


int relocation(ASM *address, char action)
{
  int i, error;
//  short relcnt_saved;
  char rom_too_old = 0;
  char kernel_too_old = 0;

  lib_name *lib_names;
  ASM *lib_contents;
  SYM_ENTRY *lib_sym;
  func_reloc_table *functions_table;
  ushort *offsets;
  ushort offset;
  ulong func_address;

  calculator = *RESET_VECTOR >= 0x400000 ? TI_92 : TI_89;

  if ((action == UNRELOC) && anticrash_running)      // run exit points
    if (address->exit_offset)
    {
  	  asm(" 	movem.l	%d0-%d7/%a0-%a6,-(%sp)");
      ASM_call((void *)address + address->exit_offset);		// Executes the _exit point
      asm("  	movem.l	(%sp)+,%d0-%d7/%a0-%a6");
    }

// LIBRARIES
  lib_names = &address->libs_names;
  error = 0;
  for (i=0; i<address->nblibs; i++)		// Boucle de recherche et de relogement des libs
  {
    int reloc_result;
    char *lib_name = lib_names[i].name;
// recherche

    lib_sym = cherche_lib(lib_name);

    if (!lib_sym)	// lib not found
    {
      error_string = lib_name;
      error = LIB_NOT_FOUND;
      break;
    }

// verifications
    lib_contents = HeapDeref(lib_sym->handle)+2;
    if ((lib_contents->sign != 0x36386B50) &&	 // 68kP
    	(lib_contents->sign != 0x36386B4C))		 // 68kL
    {
      error_string = lib_name;
      error = FILE_NOT_VALID;
      break;
    }
    if ((action == RELOC) && (lib_sym->flags.bits.archived)) // La lib est archive
    {
      ushort taille = *(((ushort*)lib_contents)-1);
      HANDLE RAM_hdl;
      taille += 2;
      RAM_hdl = HeapAllocHigh(taille);
      if (!RAM_hdl)
      {
        error = NOT_ENOUGH_MEM;
        break;
      }
      memcpy(HeapDeref(RAM_hdl), ((void *)lib_contents)-2, taille);
      lib_contents = HeapDeref(RAM_hdl)+2;

      lib_sym->compat = lib_sym->handle;
      lib_sym->handle = RAM_hdl;
      lib_sym->flags.bits.archived = 0;
      lib_sym->flags.bits.local = 1;
    }

// relogement des libs utilises

    if (action == RELOC)
      reloc_result = reloc2(lib_contents);
    else
      reloc_result = unreloc2(lib_contents);
    if (reloc_result)
    {
      error = reloc_result;
      break;
    }
  }      // Fin de la recherche des libs

  if (error)	// error = error number, and error_string is updated
  {
    if (action == RELOC)
      exit_reloc(address, i);
    return error;
  }
// Checking libs versions

  functions_table = (func_reloc_table *)&lib_names[address->nblibs];
  for (i=0; i<address->nblibs; i++)	// Pour chaque librarie
  {
    int j;
    char *lib_name = lib_names[i].name;
    ASM *lib = HeapDeref(cherche_lib(lib_name)->handle)+2;
    export_table *export = ((void *)lib + lib->export_offset);  // prend la table d'exportation de la lib

    function_table *funcs_table = &functions_table->table;

    for (j=0; j<functions_table->nbfuncs+1; j++)   // Pour chaque fonction de la lib
    {
      int index = funcs_table->index;            // indice de la fonction
	  offsets = &funcs_table->offsets;
      if (index >= export->nbfuncs)
      {
        exit_reloc(address, address->nblibs);
        error_string = lib_name;
        return WRONG_LIB_VERSION;
      }
      while ((offset = *offsets++))	// skip offsets
      { }

      funcs_table = (function_table *)offsets;
    }
    functions_table = (func_reloc_table *)funcs_table;
  }

// Relocating with libraries
  functions_table = (func_reloc_table *)&lib_names[address->nblibs];
  for (i=0; i<address->nblibs; i++)	// Pour chaque librarie
  {
    int j;
    char *lib_name = lib_names[i].name;
    export_table *export;
    ASM *lib;
    function_table *funcs_table = &functions_table->table;
    lib_sym = cherche_lib(lib_name);
    lib = HeapDeref(lib_sym->handle)+2;
	export = ((void *)lib + lib->export_offset);

    for (j=0; j<functions_table->nbfuncs+1; j++)   // Pour chaque fonction de la lib
    {
      int index = funcs_table->index;            // indice de la fonction
	  offsets = &funcs_table->offsets;
      func_address = (ulong) ((void *)lib+(&export->offsets)[index]);	// Address of the library exported function

      while ((offset = *offsets++))	// tant qu'il faut reloger qqpart
      {
        ulong *where_to_reloc = (ulong *)((void *)address + offset);
        if (action == RELOC)
          *where_to_reloc += func_address;
        else
          *where_to_reloc -= func_address;
      }
      funcs_table = (function_table *)offsets;
    }
    functions_table = (func_reloc_table *)funcs_table;
  }
  
// ROM CALLS
  if ( *((ushort *)(functions_table))++)	// si ya des rom calls
  {
    function_table *funcs_table = &functions_table->table;
    for (i=0; i<functions_table->nbfuncs+1; i++)   // Pour chaque fonction de la lib
    {
      int index = funcs_table->index;            // indice de la fonction
      int nb_rom_funcs = (*(ushort**)0xC8)[-1];

      offsets = &funcs_table->offsets;

      func_address = (*(ulong**)0xC8)[index] ;	// Address of the function
      if (index >= nb_rom_funcs)
        rom_too_old = 1;

      while ((offset = *offsets++))	// tant qu'il faut reloger qqpart
      {
        ulong *where_to_reloc = (ulong *)((void *)address + offset);
        if (action == RELOC)
          *where_to_reloc += func_address;
        else
          *where_to_reloc -= func_address;
      }
      funcs_table = (function_table *)offsets;
    }
    functions_table = (func_reloc_table *)funcs_table;
  }


// RAM CALLS
  if ( *((ushort *)(functions_table))++)	// si ya des ram calls
  {
    function_table *funcs_table = &functions_table->table;
    for (i=0; i<functions_table->nbfuncs+1; i++)   // Pour chaque fonction de la lib
    {
      int index = funcs_table->index;            // indice de la fonction
      int true_index = index & 0x3FFF;
      offsets = &funcs_table->offsets;
      if (index & EXTRA_RAM)
      {
        ushort *RAM_base = (ushort *)((void *)address+address->extraRAM_offset);
        func_address = calculator ? RAM_base[true_index*2+1] : RAM_base[true_index*2];
      }
      else
      {
		func_address = calculator ? RAM_CALLS[true_index*2+1] : RAM_CALLS[true_index*2];
        if (true_index >= NB_RAM_CALLS)
          kernel_too_old = 1;
      }

      while ((offset = *offsets++))	// tant qu'il faut reloger qqpart
      {
        ulong *where_to_reloc = (ulong *)((void *)address + offset);
        if (!index&(~WORD_REF))	// RAM CALL CALCULATOR
        {
          func_address = (ulong) &calculator;
        }
        if (action == RELOC)
        {
          if (index & WORD_REF)
            *(ushort *)where_to_reloc += (ushort)func_address;
          else
            *where_to_reloc += func_address;
        }
        else
        {
          if (index & WORD_REF)
            *(ushort *)where_to_reloc -= (ushort)func_address;
          else
            *where_to_reloc -= func_address;
        }
      }
      funcs_table = (function_table *)offsets;
    }
    functions_table = (func_reloc_table *)funcs_table;
  }
// RELOC TABLE
  func_address = (ulong) address;
  offsets = (ushort *)functions_table;
  while ((offset = *offsets++))	// tant qu'il faut reloger qqpart
  {
    ulong *where_to_reloc = (ulong *)((void *)address + offset);
    if (action == RELOC)
      *where_to_reloc += func_address;
    else
      *where_to_reloc -= func_address;
  }
// BSS
  offsets += 2;
  if (address->bss_hdl)		// si relogement BSS
  {
    func_address = (ulong) HeapDeref(address->bss_hdl);
    while ((offset = *offsets++))	// tant qu'il faut reloger qqpart
    {
      ulong *where_to_reloc = (ulong *)((void *)address + offset);
      if (action == RELOC)
        *where_to_reloc += func_address;
      else
        *where_to_reloc -= func_address;
    }
  }
  if (rom_too_old || kernel_too_old)
    return rom_too_old ? ROM_VERSION_ERROR : KERNEL_VERSION_ERROR;

  return NO_ERROR;
}